// Driver for alarm sounder
// Alarm output is OC1B

#include "typedefn.h"
#include "buttons.h"
#include "ambient.h"
#include "rtc.h"
#include "alarm.h"
#include "display.h"

#define TONEHZ     2000             // Tone frequency
#define TIMEBASE  (500000 / TONEHZ) // Half-period

static enum {OFF, READY, ACTIVE, LOCKED} AlarmState = OFF;

typedef enum                        // Reasons for sounding the tone
{
    ALARM,                          // Alarm clock
    CLICK,                          // Keyclick
    ERROR                           // Failure signal
} REASONS;

static S_TIME AlarmTime = {0};      // Loaded by main loop
static uint8_t reason   = 0;        // Reasons to sound the alarm

//-----------------------------------------------------------------------
// Local Functions
//-----------------------------------------------------------------------
static inline void ToneOn(uint8_t key)
{
    reason |= key;
    OCR1B = TIMEBASE / 2;           // Flip at mid-span
}

static inline void ToneOff(uint8_t key)
{
    reason &= ~key;
    if(reason == 0)                 // If all alarm reasons are false
    {
        OCR1B = TIMEBASE + 2;       // Never attained
    }
}

//-----------------------------------------------------------------------
// Global Functions
//-----------------------------------------------------------------------

// Initialisation
void AlarmInitialise(void)
{
    DDRB  |= BV(2);                 // OCR1B as output
    OCR1A  = TIMEBASE;              // Counter length
    ToneOff(0xff);
    TCCR1A = BV(COM1B0);            // Commit the output pin
    TCCR1B = BV(CS11) | BV(WGM12);  // 1MHz (prescaler=8), modulo=OCR1A

    AlarmState = (RtcGetAlarmMode()) ? READY : OFF;
}

//-----------------------------------------------------------------------
// Test the alarm armed-state
bool AlarmIsLive(void)
{
    switch(AlarmState)
    {
    case READY:
    case ACTIVE:
        return true;
    default:
        return false;
    }
}

//-----------------------------------------------------------------------
// Start error tone
void AlarmErrorOn(void)
{
    ToneOn(BV(ERROR));
}

//-----------------------------------------------------------------------
// Start alarm tone
void AlarmToneOn(void)
{
    ToneOn(BV(ALARM));
}

//-----------------------------------------------------------------------
// Stop alarm tone
void AlarmToneOff(void)
{
    ToneOff(BV(ALARM));
}

//-----------------------------------------------------------------------
// Start keyclick tone
void AlarmTickOn(void)
{
    ToneOn(BV(CLICK));
}

//-----------------------------------------------------------------------
// Stop keyclick tone
void AlarmTickOff(void)
{
    ToneOff(BV(CLICK));
}

//-----------------------------------------------------------------------
// Flip the alarm in/active state
void AlarmFlipState(void)
{
    switch(AlarmState)
    {
    case OFF:
        AlarmState = READY;
        AlarmToneOn();
        break;
    case READY:
        AlarmState = OFF;
        AlarmToneOff();
        break;
    default:
        return;
    }
    RtcSetAlarmMode(AlarmState == READY);
}

//-----------------------------------------------------------------------
// Set the alarm live
void AlarmSetReady(void)
{
    AlarmState = READY;     // Set the alarm active
}

//-----------------------------------------------------------------------
// Display the current alarm time (used only for setting)
void AlarmDisplayTime(void)
{
    DisplayTime(AlarmTime, ALARMSET);
}

//-----------------------------------------------------------------------
// Maintain alarm time from RTC
void AlarmReadTime(void)
{
    RtcGetAlarm(&AlarmTime);
}

//-----------------------------------------------------------------------
// Polled from the main loop: manage the alarm state
void AlarmExecute(S_TIME CurrentTime)
{
    switch(AlarmState)
    {
    case ACTIVE:                            // Alarm sounding: any button will stop it
        if(ButtonGetValue() != 0)
        {
            AlarmState = LOCKED;
            AlarmToneOff();
        }
        break;
    case READY:                                     // Is it time for the alarm?
        if(RtcTimesEqual(CurrentTime,AlarmTime))
        {                                               // Yes...
            AlarmState = ACTIVE;
            AlarmToneOn();
        }
        break;
    case LOCKED:                            // Alarm ended: don't re-arm until the time has moved on
        if(!RtcTimesEqual(CurrentTime,AlarmTime))
        {
            AlarmState = READY;
        }
        break;
    default:                                        // Alarm disabled
        ;
    }
}
