// Manage the display brightness control
// Read the ambient-light level via ADC7
// Called only by the base-level loop, which uses polled mode

#include "typedefn.h"
#include "main.h"
#include "ambient.h"
#include "rtc.h"
#include "display.h"

#define FILTEREXP 16        // Filter gain = 2^N

uint16_t FilteredAdc = 0;
uint16_t RawAdc = 0;
static S_BRIGHT limits;

//--------------------------------------------------------------
// Brightness setting by linear interpolation
#define X1 (limits.low.adc)      // Limiting values
#define X2 (limits.high.adc)
#define Y1 ((uint16_t)limits.low.timer)
#define Y2 ((uint16_t)limits.high.timer)

static inline uint16_t Interpolate(uint16_t X)
{
    return(Y1 + (((X-X1)*(Y2-Y1))/(X2-X1)));
}

//--------------------------------------------------------------
// Start up the ADC measurement subsystem
//--------------------------------------------------------------
void AmbientInitialise(void)
{
    ADMUX  = AMBIENTADC;        // Only this channel
                                // Free run, at slowest rate
    ADCSRA = BV(ADEN) | BV(ADSC) |BV(ADFR) | BV(ADIE) | BV(ADPS2) | BV(ADPS1) | BV(ADPS0);
 
    RtcGetBright(&limits);
}

//------------------------------------------------------------------------
// Set one end of the ambient-light scale
void AmbientAdjustBrightnessLimit(E_ENDS highpt, int8_t adj)
{
    S_LIMIT *pw = (highpt == HIGH) ? &limits.high : &limits.low;
 
    pw->adc   = FilteredAdc;    // New set-point
    pw->timer+= adj;

    if(pw->timer > BRIGHTMAX)         // Limits
    {
        pw->timer = BRIGHTMAX;
    }
    else if(pw->timer < BRIGHTMIN)
    {
        pw->timer = BRIGHTMIN;
    }
 
    RtcSetBright(&limits);      // Save the new value
}

//------------------------------------------------------------------------
// Adjust the display brightness
void AmbientAdjustBrightness(void)
{
    if(FilteredAdc > limits.high.adc)
    {                           // Brighter than limit
        DisplaySetBrightness(limits.high.timer);
    }
    else if(FilteredAdc < limits.low.adc)
    {                           // Dimmer than limit
        DisplaySetBrightness(limits.low.timer);
    }
    else
    {                           // Linear interpolation
        DisplaySetBrightness((uint8_t)Interpolate(FilteredAdc));
    }
}

//------------------------------------------------------------------------
// Put the current settings on the display
void AmbientDisplaySetting(E_ENDS highpt)
{
    S_LIMIT *pw = (highpt == HIGH) ? &limits.high : &limits.low;

    DisplayDecimal(pw->timer);
}

//------------------------------------------------------------------------
// Local Functions
//------------------------------------------------------------------------

//------------------------------------------------------------------------
// ADC completion interrupt: simple IIR filter (ADC runs continuously)
// Raw values: Bright sun: 800
//             Darkness:   140
ISR(SIG_ADC)
{
    static uint32_t sum = 0;	// This must maintain full accuracy

    DISABLE();
    RawAdc = ADC;
    UNDISABLE();
	
	if(FilteredAdc == 0)		// Fast-start the filter
	{
		sum = (uint32_t)RawAdc << FILTEREXP;
	}		
								// Simple IIR filter
    FilteredAdc = sum >> FILTEREXP;
    sum = sum - FilteredAdc + RawAdc;
}
