// Control-Button Handler
// Implements debounce & auto-repeatflag of the control buttons
// Buttons are timed by Timer-0, with prescaler=256
// So timer tick=8000000/(256*256)=120Hz

#include "typedefn.h"
#include "buttons.h"
#include "ambient.h"
#include "alarm.h"
#include "main.h"

#define DEBOUNCEWAIT    5       // No. of stable readings reqd.
#define AUTODELAY     120       // Periods before auto-repeat flag starts
#define AUTOREPEAT     30       // Periods between auto-repeats

static uint8_t oldbutton = 0;   // Last-seen button value
static uint8_t dlycount  = 0;   // Auto-repeat, etc., timer
 
static uint8_t BtnValue  = 0;   // Value posted to base level

//------------------------------------------------------------------------
// Global Functions
//------------------------------------------------------------------------

void ButtonInitialise(void)
{
    TCCR0  = BV(CS02);          // Prescaler = 256
    TIMSK |= BV(TOIE0);         // Enable interrupt

    oldbutton = ButtonGetRaw(); // So we don't auto-repeat the start state
    dlycount  = 0;
}

//--------------------------------------------------------------
// Return latest press, & clear it
uint8_t ButtonGetValue(void)
{
    uint8_t temp;
    DISABLE();                  // Atomic read-and-clear
    temp     = BtnValue;
    BtnValue = 0;
    UNDISABLE();
    return temp;
}

//--------------------------------------------------------------
// Static Functions
//--------------------------------------------------------------

// T0 Timer Overflow Interrupt
// This is the debounce/repeatflag timer
ISR(SIG_OVERFLOW0)
{
    uint8_t newbutton = ButtonGetRaw();

    AlarmTickOff();                         // End any keyclick

    if(newbutton == oldbutton)
    {                               // State constant: what to do...
        switch(dlycount)
        {
        case 0:                         // Inactive
            break;

        case DEBOUNCEWAIT:              // Post an event (initial press)
            BtnValue = newbutton;
            AlarmTickOn();
            dlycount++;
            break;

        case DEBOUNCEWAIT+AUTODELAY:    // Auto-repeat
            BtnValue  = newbutton | BV(REPTBIT);
            AlarmTickOn();
            dlycount -= AUTOREPEAT;     // Delay for next auto-repeat
            break;

        default:                        // Run out the timer
            dlycount++;
        }
    }
    else
    {                               // Value has changed: press or release?
        if((oldbutton & ~newbutton) != 0)
        {                               // Release: no action
            dlycount = 0;
        }
        else
        {                               // New press: start counting
            dlycount = 1;
        }
        oldbutton = newbutton;
    }
}
