/* Driver for 4-digit LED display
   Includes segment decoding
*/

#include "typedefn.h"
#include "display.h"
#include "ambient.h"
#include "main.h"
#include "rtc.h"

//--------------------------------------------------------------
// Global Data
//--------------------------------------------------------------


//--------------------------------------------------------------
// Static data
//--------------------------------------------------------------
// Digit selectors: NB active low
#define DGT1N   0       // Port D
#define DGT2N   1
#define DGT3N   0       // Port B
#define DGT4N   1

// Encoding of a display digit. For PCB layout reasons, the lines are
// spread across all of Ports B, C & D. As there are no other direct
// output lines (all outputs are from specific modules), we can simply
// write values direct to the ports, with no need to mask bits.

// Bits assigned to segments, in 3 ports
#define SA 2    // Port D
#define SB 3
#define SC 4
#define SD 6    // Port B
#define SE 7
#define SF 5    // Port D
#define SG 6
#define DL 7
#define DR 0    // Port C

// DDR values (all output bits)
#define DDB (BV(SD) | BV(SE) | BV(DGT3N) | BV(DGT4N))
#define DDC (BV(DR))
#define DDD (BV(SA) | BV(SB) | BV(SC) | BV(SF) | BV(SG) | BV(DL) | BV(DGT1N) | BV(DGT2N))

// Offsets in map of 3 output ports
#define PB  0
#define PC  8
#define PD 16

// Segment bits, in their OP ports
#define SEGA BVL(SA+PD)
#define SEGB BVL(SB+PD)
#define SEGC BVL(SC+PD)
#define SEGD BVL(SD+PB)
#define SEGE BVL(SE+PB)
#define SEGF BVL(SF+PD)
#define SEGG BVL(SG+PD)
#define DPL  BVL(DL+PD)
#define DPR  BVL(DR+PC)

#define HALFSEC     30              // Divisor for the 2Hz timer
#define UPDATESEC   10              // RTC update period (seconds)

typedef enum        // Indices into the following table
{
    ZERO = 0,
    ONE,
    TWO,
    THREE,
    FOUR,
    FIVE,
    SIX,
    SEVEN,
    EIGHT,
    NINE,
    BLANK,
    DASH,
    ERROR
} E_VALUE;

// The BCD - 7-segment output table, in ROM
static const uint32_t Decode[] PROGMEM = {
    SEGA + SEGB + SEGC + SEGD + SEGE + SEGF       ,   // 0
           SEGB + SEGC                            ,   // 1
    SEGA + SEGB        + SEGD + SEGE        + SEGG,   // 2
    SEGA + SEGB + SEGC + SEGD               + SEGG,   // 3
           SEGB + SEGC               + SEGF + SEGG,   // 4
    SEGA        + SEGC + SEGD        + SEGF + SEGG,   // 5
    SEGA        + SEGC + SEGD + SEGE + SEGF + SEGG,   // 6
    SEGA + SEGB + SEGC                            ,   // 7
    SEGA + SEGB + SEGC + SEGD + SEGE + SEGF + SEGG,   // 8
    SEGA + SEGB + SEGC + SEGD        + SEGF + SEGG,   // 9
    0                                             ,   // Blank
                                              SEGG,   // -
    SEGA               + SEGD + SEGE + SEGF + SEGG    // E
};

static uint32_t DisplayData[4] = {0};       // Data to be displayed
static bool     Hrs24          = false;     // 24-hour display mode
static bool     HalfSecondFlag = false;     // Flipped each half-second
static bool     UpdateFlag     = true;      // RTC update-display flag

static void AllDigitsOff(void);

//--------------------------------------------------------------
// Global Functions
//--------------------------------------------------------------

// Check if we need to read the time
bool DisplayTimeToUpdate(void)
{
    bool upflag;

    DISABLE();
    upflag = UpdateFlag;        // Atomic test-&-clear
    UpdateFlag = false;
    UNDISABLE();
    return upflag;
}

// Convert digit/char to 7-segment (over 3 IO ports)
uint32_t DisplayBcdTo7Seg(uint8_t digit)
{
    switch(digit)               // Map value to table index
    {
    case 0 ... 9:               // BCD digits map directly
        break;
    case ' ':
        digit = BLANK;
        break;
    case '-':
        digit = DASH;
        break;
    default:
        digit = ERROR;
        break;
    }
    return pgm_read_dword_near(&Decode[digit]); // Segment data
}

//--------------------------------------------------------------
// Toggle the 12/24-hr display mode
void DisplayFlipMode(void)
{
    Hrs24 = !Hrs24;
    RtcSetDisplayMode24(Hrs24);
}

//--------------------------------------------------------------

// Set the display brightness, as an 8-bit value
void DisplaySetBrightness(uint8_t value)
{
    if(value < BRIGHTMIN)
    {
        value = BRIGHTMIN;
    }
    else
    if(value > BRIGHTMAX)
    {
        value = BRIGHTMAX;
    }

    OCR2 = value;               // Comparator value
}

//--------------------------------------------------------------
// Initialise the display timer (T2)
// The timing is 8MHz(osc) / 128(prescaler) / 256(counter) = 244Hz / 4(digits) = 60Hz(panel rate).
// A counter from panel rate generates a 2Hz (nominal) timer for base-level actions.
void DisplayInitialise(void)
{
    uint8_t digit;
 
    AllDigitsOff();

    DDRB |= DDB;                            // Port data-direction registers
    DDRC |= DDC;
    DDRD |= DDD;

    for(digit=0; digit<4; digit++)
    {
        DisplayData[digit] = 0;             // Blank the display
    }
    Hrs24 = RtcIs24HrDisplay();
 
    ASSR &= ~BV(AS2);                       // Timer 2 in synchronous mode
    TCCR2 = BV(CS22) | BV(CS20);            // Prescaler = 128, start timer, modulo-256
    DisplaySetBrightness(BRIGHTMAX);        // Max. brightness, initially
 
    TIMSK |= (BV(OCIE2) | BV(TOIE2));       // Interrupt on compare or overflow
}

//--------------------------------------------------------------
// Show the time as given
void DisplayTime(S_TIME time, E_DISPMODE mode)
{
    uint8_t  rawtime[4];
    uint32_t rawdisplay[4] = {0};
    uint8_t  ctr;
 
    rawtime[0] = time.hours >> 4;			// Convert from BCD
    rawtime[1] = time.hours & 0x0f;
    rawtime[2] = time.minutes >> 4;
    rawtime[3] = time.minutes & 0x0f;

    ctr = (rawtime[0] * 10) + rawtime[1];   // Hours, 24-notation, in binary
 
    if(Hrs24 == false)						// Convert 24-hour to 12-hour notation
    {
        switch(ctr)
        {
        case 0:
            rawtime[0] = 1;					// No zero in 12-hour display (after midnight)
            rawtime[1] = 2;					// 00 -> 12
            break;

        case 13 ... 23:						// Fold back into 12-hours
            ctr -= 12;
            rawtime[0] = ctr / 10;
            rawtime[1] = ctr % 10;			// Fall through

        case 12:							// After midday
            rawdisplay[0] |= DPL;			// The "PM" flag
            break;

        default:							// 1..11 unchanged
            ;
        }

        if(rawtime[0] == 0)					// Suppress leading zero
        {
            rawtime[0] = ' ';
        }
    }

    switch(mode)						// Set the DPs as per the mode
    {
    case ALARMSET:							// Alarm is live
        rawdisplay[3] |= DPR;				// Fall through

    case MAIN:
        if(HalfSecondFlag)					// Normal: cycle 2 dots
        {
            rawdisplay[1] |= DPR;
        }
        else
        {
            rawdisplay[2] |= DPL;
        }
        break;
    case MAINSET:
        break;          // No DP here
    }
 
    for(ctr=0; ctr<4; ctr++)
    {
        DisplayData[ctr] = rawdisplay[ctr] | DisplayBcdTo7Seg(rawtime[ctr]);
    }
}

//--------------------------------------------------------------
// Get & display the RTC calibration
void DisplayCalibration(void)
{
    uint8_t rawcal = RtcGetCalibration();

    // Convert from sign/magnitude form
    DisplayDecimal(rawcal & 0x7f);
    DisplayData[0] = DisplayBcdTo7Seg((rawcal & 0x80) ? '-' : ' ');
}

//--------------------------------------------------------------
// Display a number 0 - 9999 right justified (no decimal pts.)
void DisplayDecimal(uint16_t val)
{
    int8_t  i  = 3;
    uint8_t lz = 0;

    if(val > 9999)
    {                           // Out of range
        lz = 'E';
    }
    else
    {
        for(; val > 0; i--)
        {
            uint8_t r = val % 10;
 
            val /= 10;
            DisplayData[i] = DisplayBcdTo7Seg(r);
            lz = ' ';
        }
    }
    while(i >= 0)
    {                           // Blank leading zeros
        DisplayData[i--] = DisplayBcdTo7Seg(lz);
        lz = ' ';
    }
}

//--------------------------------------------------------------
// Static Functions
//--------------------------------------------------------------

//--------------------------------------------------------------
// Turn all 4 digits off. Coded so GCC can use the AVR's fast IO bit instructions
// NB The digit drives are active low
static void AllDigitsOff(void)
{
    PORTB &= ~(BV(SD) | BV(SE));    // All segments off
    PORTC &= ~(BV(DR));
    PORTD &= ~(BV(SA) | BV(SB) | BV(SC) | BV(SF) | BV(SG) | BV(DL));

    PORTD |= BV(DGT1N);             // Drive all the digits inactive
    PORTD |= BV(DGT2N);             // Don't merge these: there are fast 1-bit instructions
    PORTB |= BV(DGT3N);
    PORTB |= BV(DGT4N);
}

//--------------------------------------------------------------

// T2 Timer Overflow Interrupt
// This is the signal to switch to the next digit
ISR(SIG_OVERFLOW2)
{
    static uint8_t digit   = 0;             // Digit to display
    static uint8_t halfsec = 0;             // Prescaler to nominal half-seconds
    static uint8_t upctr   = 0;             // Main display-update counter
 
    AllDigitsOff();                         // Be sure they are off

    if(digit >= 3)                          // Wrap over
    {
        digit = 0;
        if(halfsec >= HALFSEC)              // Half-second point
        {
            halfsec = 0;
            HalfSecondFlag = !HalfSecondFlag;

            if(upctr >= UPDATESEC)
            {                               // Request a reload of current time
                upctr = 0;
                UpdateFlag = true;
            }
            else
            {
                upctr++;
            }
        }
        else
        {
            halfsec++;
        }
    }
    else
    {
        digit++;                            // Next digit
    }

    PORTB = ((DisplayData[digit] >> PB) & 0xFF) | BV(DGT3N) | BV(DGT4N);
    PORTC =  (DisplayData[digit] >> PC) & 0xFF;
    PORTD = ((DisplayData[digit] >> PD) & 0xFF) | BV(DGT1N) | BV(DGT2N);

    switch(digit)                           // Turn on the reqd. digit
    {
    case 0:
        PORTD &= ~BV(DGT1N);
        break;
    case 1:
        PORTD &= ~BV(DGT2N);
        break;
    case 2:
        PORTB &= ~BV(DGT3N);
        break;
    default:
        PORTB &= ~BV(DGT4N);
        break;
    }
}

//--------------------------------------------------------------

// T2 Timer Compare Interrupt
// This is used by the brightness control: it simply turns the display off
// until the next digit time.
// This interrupt has priority over the overflow interrupt.
ISR(SIG_OUTPUT_COMPARE2)
{
    AllDigitsOff();
}
