/* I2C polled-mode driver
   Here, used only for the RTC device.
   Basic read & write functions
*/

#include "typedefn.h"
#include "i2c.h"
#include "alarm.h"

#define READOP  1               // Read operation
#define WRITEOP 0               // Write operation

// TWSR Status Codes (Master mode only)
#define TWSR_ONEST      0x08    // A START condition has been transmitted
#define TWSR_REPST      0x10    // A repeated START condition has been transmitted
#define TWSR_SLAWACK    0x18    // SLA+W has been transmitted; ACK has been received
#define TWSR_SLAWNAK    0x20    // SLA+W has been transmitted; NOT ACK has been received
#define TWSR_DATTACK    0x28    // Data byte has been transmitted; ACK has been received
#define TWSR_DATTNAK    0x30    // Data byte has been transmitted; NOT ACK has been received
#define TWSR_ARBLOST    0x38    // Arbitration lost in SLA+W or data bytes
#define TWSR_SLARACK    0x40    // SLA+R has been transmitted; ACK has been received
#define TWSR_SLARNAK    0x48    // SLA+R has been transmitted; NOT ACK has been received
#define TWSR_DATRACK    0x50    // Data byte has been received; ACK has been returned
#define TWSR_DATRNAK    0x58    // Data byte has been received; NOT ACK has been returned

static void RtcPause(uint8_t expect);
static void RtcBegin(uint8_t slave, uint8_t addr);

//---------------------------------------------------------------------
// Global Functions
//---------------------------------------------------------------------

// Initialise the I2C channel, in polled mode

void I2CInitialise(void)
{
    TWSR = 0;                   // TWPS = 00 - prescaler = 1
    TWBR = 32;                  // Bit clock = 100kHz
    TWCR = BV(TWEN);            // Enable the TWI (aka I2C) bus
}

//---------------------------------------------------------------------

// Read a block from the I2C device

bool I2CReadBlock(uint8_t slave, uint8_t addr, uint8_t length, void *buffpt)
{
    uint8_t *buff = (uint8_t*)buffpt;
    RtcBegin(slave, addr);
    TWCR = BV(TWINT) | BV(TWSTA) | BV(TWEN);    // Send Repeated START condition
    RtcPause(TWSR_REPST);
    TWDR = slave + READOP;
    TWCR = BV(TWINT) | BV(TWEN);                // RTC address & Read command
    RtcPause(TWSR_SLARACK);
    for(; length>1; length--)                   // All bytes except the last
    {
        TWCR = BV(TWINT) | BV(TWEN) | BV(TWEA);
        RtcPause(TWSR_DATRACK);
        *(buff++) = TWDR;                       // Data as read
    }

    TWCR = BV(TWINT) | BV(TWEN);                // No ACK on last/only byte
    RtcPause(TWSR_DATRNAK);
    *(buff++) = TWDR;                           // Data as read

    TWCR = BV(TWINT) | BV(TWEN) | BV(TWSTO);    // Send STOP condition
    // No interrupt follows this
    return true;
}

//---------------------------------------------------------------------

// Write a block to the I2C device

bool I2CWriteBlock(uint8_t slave, uint8_t addr, uint8_t length, void *buffpt)
{
    uint8_t *buff = (uint8_t*)buffpt;
    RtcBegin(slave, addr);
    for(; length>0; length--)
    {
        TWDR = *(buff++);                       // Data to write
        TWCR = BV(TWINT) | BV(TWEN);
        RtcPause(TWSR_DATTACK);
    }
    TWCR = BV(TWINT) | BV(TWEN) | BV(TWSTO);    // Send STOP condition
    // No interrupt follows this
    return true;
}

//---------------------------------------------------------------------
// Local Functions
//---------------------------------------------------------------------

// Wait for I2C to pass 1 byte, check status

static void RtcPause(uint8_t expect)
{
    while(!(TWCR & BV(TWINT)))
        ;

    if((TWSR & 0xF8) != expect)
    {
        AlarmErrorOn();             // Error action, if reqd.
    }
}

//---------------------------------------------------------------------
// Start communications, send read/write address to RTC

static void RtcBegin(uint8_t slave, uint8_t addr)
{
    TWCR = BV(TWINT) | BV(TWSTA) | BV(TWEN);    // Send START condition
    RtcPause(TWSR_ONEST);
    TWDR = slave + WRITEOP;
    TWCR = BV(TWINT) | BV(TWEN);                // RTC address & Write command
    RtcPause(TWSR_SLAWACK);
    TWDR = addr;
    TWCR = BV(TWINT) | BV(TWEN);                // Address to write to
    RtcPause(TWSR_DATTACK);
}
