// Simple digital clock project
// Mainline

/*
    Security/Configuration Bits:
    PonyProg:	AVR Studio:
    [SPIEN]		RSTDISBL
    BOOTSZ1		SPIEN		Y
    BOOTSZ0		EESAVE
    SUT0		BOOTSZ		1024W_0C00
    CKSEL3		BOOTRST
    CKSEL1		CKOPT
    CKSEL0		BODLEVEL	2V7
				BODEN
				SUT_CKSEL	INTRCOSC_8MHZ_6CK_64MS
				
	Fuse HI=D9, LO=E4
	LOCKBIT=FF  (all off)
*/

#include "typedefn.h"
#include "alarm.h"
#include "ambient.h"
#include "display.h"
#include "i2c.h"
#include "rtc.h"
#include "buttons.h"
#include "main.h"

typedef enum {ShowTime, AlarmSet, TimeSet, HiBrite, LoBrite, Calibrate} E_STATE;

static E_STATE StartingValue(void);
static E_STATE MainTimeDisplay(E_STATE current);
static E_STATE MainTimeSet(E_STATE current);
static E_STATE AlarmTimeSet(E_STATE current);
static E_STATE CalibrateRtc(E_STATE current);
static E_STATE MainSetBrightness(bool highpt, E_STATE current);

extern uint16_t FilteredAdc;

//------------------------------------------------------------------------
// Global Functions
//------------------------------------------------------------------------

// Project mainline
int main(void)
{
    E_STATE State;

    I2CInitialise();                // This must happen first
    RtcInitialise();
    ButtonInitialise();
                                    // Following fns. get initz. data from RTC
    AlarmInitialise();
    DisplayInitialise();
    AmbientInitialise();

    State = StartingValue();        // Buttons at power-up
 
    _WDR();							// Clear watchdog counter
    WDTCR |= BV(WDCE) | BV(WDE);    // Enable watchdog access
    WDTCR  = BV(WDE)  | 2;          // Armed, timeout = 65mS

    ENABLE();                       // Interrupts on

    for(;;)                         // The main action loop
    {
        _WDR();						// Reset the watchdog
		
        switch(State)
        {
        case ShowTime:              // Main time-display mode
            State = MainTimeDisplay(State);
            break;
 
        case AlarmSet:              // Change the alarm time
            AlarmReadTime();
            State = AlarmTimeSet(State);
            break;
 
        case TimeSet:               // Change the main time
            State = MainTimeSet(State);
            break;
 
        case HiBrite:               // Change the high-point brightness
            State = MainSetBrightness(HIGH, State);
            break;
 
        case LoBrite:               // Change the low-point brightness
            State = MainSetBrightness(LOW, State);
            break;
 
        case Calibrate:             // Change the oscillator calibration
            State = CalibrateRtc(State);
            break;
        }
        AmbientAdjustBrightness();
    }
 
    return 0;                       // Never hit!
}

//------------------------------------------------------------------------
// Local Functions
//------------------------------------------------------------------------

// Actions on power-up buttons
static E_STATE StartingValue(void)
{
    switch(ButtonGetRaw())
    {
    case BV(BUTTONA):
        return TimeSet;
    case BV(BUTTONB):
        return HiBrite;
    case BV(BUTTONC):
        return LoBrite;
    case BV(BUTTONB) | BV(BUTTONC):
        return Calibrate;
    default:
        return ShowTime;
    }
}

//------------------------------------------------------------------------
static E_STATE MainTimeDisplay(E_STATE current)
{
    static S_TIME CurrentTime;
 
    if(DisplayTimeToUpdate())   // Every few seconds, read the RTC
    {
        CurrentTime = RtcGetTime();
    }

    DisplayTime(CurrentTime, AlarmIsLive() ? ALARMSET : MAIN);
    AlarmExecute(CurrentTime);
 
    switch(ButtonGetValue())    // Disregard auto-repeat signals
    {
    case BV(BUTTONA):           // Button A: set the alarm time
        current = AlarmSet;
        break;
    case BV(BUTTONB):           // Button B: en/disable the alarm
        AlarmFlipState();
        break;
    case BV(BUTTONC):           // Button C: flip 12/24hrs mode
        DisplayFlipMode();
        break;
    default:
        ;
    }
    return current;
}

//------------------------------------------------------------------------
static E_STATE MainTimeSet(E_STATE current)
{
    DisplayTime(RtcGetTime(), MAINSET);

    switch(ButtonNoRepeat(ButtonGetValue()))
    {
    case BV(BUTTONA):           // Exit set-mode
        current = ShowTime;     // Main display active
        break;
    case BV(BUTTONC):           // Advance hours
        RtcAdvanceMainHours();
        break;
    case BV(BUTTONB):           // Advance minutes
        RtcAdvanceMainMinutes();
        break;
    }
    return current;
}

//------------------------------------------------------------------------
static E_STATE AlarmTimeSet(E_STATE current)
{
    AlarmDisplayTime();
    switch(ButtonNoRepeat(ButtonGetValue()))
    {
    case BV(BUTTONA):           // Exit set-mode, leave armed
        AlarmSetReady();
        current = ShowTime;
        break;
    case BV(BUTTONC):           // Advance hours
        RtcAdvanceAlarmHours();
        break;
    case BV(BUTTONB):           // Advance minutes
        RtcAdvanceAlarmMinutes();
        break;
    }
    return current;
}

//------------------------------------------------------------------------
static E_STATE CalibrateRtc(E_STATE current)
{
    DisplayCalibration();
    switch(ButtonNoRepeat(ButtonGetValue()))
    {
    case BV(BUTTONA):           // Exit set-mode
        current = ShowTime;     // Main display active
        break;
    case BV(BUTTONB):           // Increase
        RtcCalibrate(1);
        break;
    case BV(BUTTONC):           // Decrease
        RtcCalibrate(-1);
        break;
    }
    return current;
}

//------------------------------------------------------------------------
static E_STATE MainSetBrightness(bool highpt, E_STATE current)
{
    AmbientDisplaySetting(highpt);

    switch(ButtonNoRepeat(ButtonGetValue()))
    {
    case BV(BUTTONA):           // Exit set-mode
        current = ShowTime;     // Main display active
        break;
    case BV(BUTTONB):           // Increase
        AmbientAdjustBrightnessLimit(highpt, 1);
        break;
    case BV(BUTTONC):           // Decrease
        AmbientAdjustBrightnessLimit(highpt, -1);
        break;
    }
    return current;
}
