// Management of the MCP7940N RTC device
// NB The alarm function is not used: it is pure software

#include "typedefn.h"
#include "main.h"
#include "i2c.h"
#include "ambient.h"
#include "rtc.h"

#define RTCADDR 0xDE            // RTC chip's I2C address

typedef enum                // RTC byte addresses
{
    RTC_SECONDS = 0,            // Current time/date
    RTC_MINUTES,
    RTC_HOURS,
    RTC_DAY,
    RTC_DATE,
    RTC_MONTH,
    RTC_YEAR,
    RTC_CONTROL,
 
    RTC_CALIBN,
    RTC_RESERVED1,
    RTC_ALM0SECONDS,            // Alarm 0
    RTC_ALM0MINUTES,
    RTC_ALM0HOURS,
    RTC_ALM0DAY,
    RTC_ALM0DATE,
    RTC_ALM0MONTH,
 
    RTC_RESERVED2,
    RTC_ALM1SECONDS,            // Alarm 1
    RTC_ALM1MINUTES,
    RTC_ALM1HOURS,
    RTC_ALM1DAY,
    RTC_ALM1DATE,
    RTC_ALM1MONTH,
    RTC_RESERVED3,
 
    RTC_TS0MINUTES,             // Timestamp 0 (power fails)
    RTC_TS0HOURS,
    RTC_TS0DATE,
    RTC_TS0MONTH,
    RTC_TS1MINUTES,             // Timestamp 1 (power returns)
    RTC_TS1HOURS,
    RTC_TS1DATE,
    RTC_TS1MONTH,
                            // RTC RAM area - software use
    RTC_BRIGHT,                 // sizeof(S_BRIGHT)
    RTC_BRIGHT1,
    RTC_BRIGHT2,
    RTC_BRIGHT3,
    RTC_BRIGHT4,
    RTC_BRIGHT5,
    RTC_SALMMINS,               // Software alarm: minutes
    RTC_SALMHOURS,              //                 hours
    RTC_24HRS,                  // TRUE for 24-hour display
    RTC_ALARMLIVE,              // Software alarm is active
} E_RTC;

                            // RTC_SECONDS
#define ST      7               // Start flag
                            // RTC_DAY
#define VBATEN  3               // Backup power enable
#define VBAT    4               // Set by RTC on power loss
#define OSCON   5               // Set by RTC when oscillator is live

static bcdigs RtcIncrement(bcdigs val, bcdigs limit);
static void   RtcColdStart(void);
static bool   RtcIsRunning(void);

//---------------------------------------------------------------------
// Global Functions
//---------------------------------------------------------------------
// Start up the RTC interface (NB This is not the RTC cold start)
// NB The I2C driver must have been already started
void RtcInitialise(void)
{
    if(!RtcIsRunning())             // Start the RTC if needed
    {
        RtcColdStart();
    }
}

//---------------------------------------------------------------------

// Get the ALARM-enable state
bool RtcGetAlarmMode(void)
{
    bool val;

    I2CReadBlock(RTCADDR, RTC_ALARMLIVE, sizeof(val), &val);
    return (val != 0);
}

//---------------------------------------------------------------------

// Set the software alarm enable
void RtcSetAlarmMode(bool val)
{
    I2CWriteBlock(RTCADDR, RTC_ALARMLIVE, sizeof(val), &val);
}

//---------------------------------------------------------------------

// Return TRUE if 24-hour display is selected
bool RtcIs24HrDisplay(void)
{
    uint8_t val;
 
    I2CReadBlock(RTCADDR, RTC_24HRS, sizeof(val), &val);
    return (val != 0);
}

//---------------------------------------------------------------------

// Set the display mode
void RtcSetDisplayMode24(bool hrs24)
{
    I2CWriteBlock(RTCADDR, RTC_24HRS, sizeof(hrs24), &hrs24);
}

//---------------------------------------------------------------------

// Set current time
void RtcSetTime(S_TIME val)
{
    uint8_t secs;
    do                          // Don't update while the RTC is about to do so
    {
        I2CReadBlock(RTCADDR, RTC_SECONDS, sizeof(secs), &secs);
    } while((secs & 0x7f) == 0x59);

    I2CWriteBlock(RTCADDR, RTC_MINUTES, sizeof(val), &val);
}

//---------------------------------------------------------------------

// Get current time
S_TIME RtcGetTime(void)
{
    S_TIME val;
    I2CReadBlock(RTCADDR, RTC_MINUTES, sizeof(val), &val);
    return val;
}

//---------------------------------------------------------------------

// Save current alarm time (NB alarm is managed by software)
// No need to wait if RTC is updating
void RtcSetAlarm(S_TIME *val)
{
    I2CWriteBlock(RTCADDR, RTC_SALMMINS, sizeof(S_TIME), val);
}

//---------------------------------------------------------------------

// Get current alarm time
void RtcGetAlarm(S_TIME *val)
{
    I2CReadBlock(RTCADDR, RTC_SALMMINS, sizeof(S_TIME), val);
}

//---------------------------------------------------------------------

// Save current brightness limits
// No need to wait if RTC is updating
void RtcSetBright(S_BRIGHT *val)
{
    I2CWriteBlock(RTCADDR, RTC_BRIGHT, sizeof(S_BRIGHT), val);
}

//---------------------------------------------------------------------

// Get current brightness limits
void RtcGetBright(S_BRIGHT *val)
{
    I2CReadBlock(RTCADDR, RTC_BRIGHT, sizeof(S_BRIGHT), val);
}

//---------------------------------------------------------------------

// Test if 2 times are equal
bool RtcTimesEqual(S_TIME a, S_TIME b)
{
    return ((a.minutes == b.minutes) && (a.hours == b.hours));
}

//------------------------------------------------------------------------
// Adjust the displayed hours, & store in RTC
void RtcAdvanceMainHours(void)
{
    S_TIME val = RtcGetTime();
    val.hours  = RtcIncrement(val.hours,0x24);
    RtcSetTime(val);
}

//------------------------------------------------------------------------
// Adjust the displayed minutes, & store in RTC
void RtcAdvanceMainMinutes(void)
{
    S_TIME val  = RtcGetTime();
    val.minutes = RtcIncrement(val.minutes, 0x60);
    RtcSetTime(val);
}

//------------------------------------------------------------------------
// Adjust the displayed hours, & store in RTC
void RtcAdvanceAlarmHours(void)
{
    S_TIME val;
    RtcGetAlarm(&val);
    val.hours  = RtcIncrement(val.hours,0x24);
    RtcSetAlarm(&val);
}

//------------------------------------------------------------------------
// Adjust the displayed minutes, & store in RTC
void RtcAdvanceAlarmMinutes(void)
{
    S_TIME val;
    RtcGetAlarm(&val);
    val.minutes = RtcIncrement(val.minutes, 0x60);
    RtcSetAlarm(&val);
}

//------------------------------------------------------------------------
// Adjust the RTC oscillator calibration
void RtcCalibrate(int8_t incr)
{
    uint8_t rawcal;             // Raw calibration, sign-magnitude form
    int16_t scal;				// Signed calibration
	
    I2CReadBlock(RTCADDR, RTC_CALIBN, sizeof(rawcal), &rawcal);
 
    if(rawcal & 0x80)           // Sign set, ie -ve
    {
        scal = -(rawcal & 0x7f);
    }
    else                        // +ve
    {
        scal = rawcal;
    }
 
    scal += incr;               // Apply the adjustment
 
    if(scal > 127)              // Limit it
    {
        scal = 127;
    }
    else if(scal < -127)
    {
        scal = -127;
    }
 
    if(scal < 0)				// Convert back to sign/magnitude form
    {
        rawcal = (uint8_t)(-scal) | 0x80;
    }
    else
    {
        rawcal = (uint8_t)scal;
    }
    I2CWriteBlock(RTCADDR, RTC_CALIBN, sizeof(rawcal), &rawcal);
}

//------------------------------------------------------------------------
// Read the current calibration
uint8_t  RtcGetCalibration(void)
{
    uint8_t val;
    I2CReadBlock(RTCADDR, RTC_CALIBN, sizeof(val), &val);
    return val;     // NB Sign-magnitude form
}

//------------------------------------------------------------------------
// Local Functions
//------------------------------------------------------------------------

//------------------------------------------------------------------------
// Increment a BCD value, wrapping at given limit
static bcdigs RtcIncrement(bcdigs val, bcdigs limit)
{
    uint8_t lsdig = val & 0x0f;
    uint8_t msdig = (val >> 4) & 0x0f;
    uint8_t ans;

    if(msdig > 5)
    {
        msdig = 0;      // Force back into range
    }

    switch(lsdig)
    {
    case 0 ... 8:       // Count up
        lsdig++;
        break;
    case 9:             // Roll over 10's
        lsdig = 0;
        msdig++;
        break;
    default:            // Invalid: force into range
        lsdig = 0;
    }

    ans = (msdig << 4) + lsdig;
    return (ans < limit) ? ans : 0;
}

//---------------------------------------------------------------------
// Test if the RTC oscillator is running
// This also proves it has not lost backup power
static bool RtcIsRunning(void)
{
    uint8_t rtcstat;
 
    I2CReadBlock(RTCADDR, RTC_DAY, sizeof(rtcstat), &rtcstat);
 
    return ((rtcstat & BV(OSCON)) != 0);
}

//---------------------------------------------------------------------

// Cold-start the RTC counter hardware after a power loss
// This sets the stored time to 00:00:00, 12-hour display
static void RtcColdStart(void)
{
    uint8_t  setrtc[4] = {BV(ST), 0, 0, BV(VBATEN)+1};
    S_BRIGHT init = {{0, 240}, {1023, 240}};
    S_TIME   alrm = {0,0};
 
    I2CWriteBlock(RTCADDR, RTC_SECONDS, sizeof(setrtc), setrtc);
    RtcSetBright(&init);
    RtcSetAlarm(&alrm);
    RtcSetAlarmMode(false);
    RtcSetDisplayMode24(false);
}
